/*
    v0.01 - original
    v0.02 - removed KEY REMAPPING
    v0.03 - changed patch.  The original would also disable the now/next on some channels.

    a tap to disable/enable mheg
    this merely prevents the software from processing the dsmcc stream
    by zapping the subrouting call.
    Note, since you can enable/disable it on the fly; you can play tricks like:
        enable mheg
        tune to a chennel
        wait a bit to allow the cache to fill up
        disable mheg
        viewing the (now static) pages
    The cache will be deleted when you change channels.

    Use the 'Recall' key to bring up a menu (exit cancels, recall propagates to next tap/firmware).

*/

#include <tap.h>

//Every TAP application needs unique ID and TAP allocations that have same ID cant execute at same time.
//ID consists of DWORD( 4byte ) and the MSB bit should be set by 1.
//The following is how to use in source codes.
#define ID_USERTAP 0xbdb00103
TAP_ID( ID_USERTAP );

//Every TAP application has to include Program name, Author name, and Program description as following
//manners in source code.
TAP_PROGRAM_NAME("mheg_off");
TAP_AUTHOR_NAME("bdb");
TAP_DESCRIPTION("A TAP to disable/enable MHEG");
TAP_ETCINFO(__DATE__);

//----------------------------------------------------------------------------
enum {
    VERSION_1209 = 0,
    VERSION_1205,
    VERSION_1204,
    NUM_VERSIONS
};
int firmware_id;

//----------------------------------------------------------------------------
#define pSysID                      0xa3fffffa
#define ptrfix(p) (p + _tap_startAddr)

extern int _appl_version;

typedef struct {
    int  id;
    char str[0x20];
}type_item;

typedef struct {
    int  id;
    int  visible;
    TYPE_Window win;
    int  rgn;
    char *fixme_titlep;
    type_item item[0x10];
}type_menu;

//----------------------------------------------------------------------------
type_menu menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {3, "xxx",          },
        {0, "Disable MHEG", },
        {1, "Enable MHEG",  },
        {2, "Exit Tap",     },
    },
};


int TAP_Main(void);
dword TAP_EventHandler(word event, dword param1, dword param2);
int menu_handler(type_menu *menu, dword param1);
int key_handler(dword param1);
void ShowMenu(type_menu *menu);
void DeleteMenu(type_menu *menu);

int patch_dsmcc();
int unpatch_dsmcc();

//----------------------------------------------------------------------------
int TAP_Main(void)
{
    int ret;
    word SysID = *(word*)pSysID;

    TAP_Print("\nopening %s",__tap_program_name__);

    switch(SysID) {
        case 456:
            break;
        default:
            TAP_Print("\nUnsupported system id - %d", SysID);
            TAP_Print("\nTAP only supports system id 456");
            TAP_Print("\n");
            return 0;
    }
    switch(_appl_version) {
        case 0x1204:
            firmware_id = VERSION_1204;
            break;
        case 0x1205:
            firmware_id = VERSION_1205;
            break;
        case 0x1209:
            firmware_id = VERSION_1209;
            break;
        default:
            TAP_Print("\nUnsupported firmware version - %x", _appl_version);
            TAP_Print("\nTAP only supports firmware versions 5.12.04,5.12.05,5.12.09");
            TAP_Print("\n");
            return 0;
    }

    menu.visible = 0;
    return 1;                 //0=application mode, 1=tsr mode
}

//----------------------------------------------------------------------------
dword TAP_EventHandler(word event, dword param1, dword param2)
{
    int ret;
    dword state, subState;

    switch (event) {
        case EVT_IDLE:
            break;

        case EVT_KEY:
//TAP_Print("\nEVT_KEY, event=%x, param1=%x, param2=%x", event, param1, param2);

            TAP_GetState(&state, &subState);
            if ((state == STATE_Normal) && (subState == SUBSTATE_Normal)) {
                if (menu.visible) { // handling keys when options window is shown
                    param1 = menu_handler(&menu, param1);
                }
                else {
                    param1 = key_handler(param1);
                }
            }
	        break;
    }
    return param1;
}

int menu_handler(type_menu *menu, dword param1)
{
    int selection;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
			switch (selection) {
				case 0:
				    patch_dsmcc();
				    return 0;
				case 1:
				    unpatch_dsmcc();
				    return 0;
				case 2:
				    unpatch_dsmcc();
                    TAP_Print("\nclosing %s",__tap_program_name__);
				    TAP_Exit();
				    return 0;
			}
            break;
        case RKEY_Recall:
			DeleteMenu(menu);
            return param1; // cascades key to other TAP's
        case RKEY_Exit:
            DeleteMenu(menu);
            return 0; // exit without Cascading
        default:
	        TAP_Win_Action(&menu->win, param1); // send all other key's to menu-Win
    		return 0;
    }
    return param1;
}


int key_handler(dword param1)
{
    switch(param1) {
        case RKEY_Recall:                 //menu
            ShowMenu(&menu);
            return 0;
    }
    return param1;
}

//----------------------------------------------------------------------------
void ShowMenu(type_menu *menu)
{
    int i;

	if (!menu->visible) {
	    menu->rgn = TAP_Osd_Create(0, 0, 720, 576, 0, FALSE);
	    TAP_Win_SetDefaultColor(&menu->win);
	    TAP_Win_Create(&menu->win, menu->rgn, 250, 210, 200, 80, FALSE, FALSE);
        TAP_Win_SetTitle(&menu->win, ptrfix(menu->fixme_titlep), 0, FNT_Size_1622);
	    for (i=1; i<=menu->item[0].id; i++) {
    	    TAP_Win_AddItem(&menu->win, menu->item[i].str);
	    }
	    TAP_Win_SetSelection(&menu->win, 0);
	    menu->visible = 1;
	}
}

void DeleteMenu(type_menu *menu)
{
	if (menu->visible) {
	    TAP_Win_Delete(&menu->win);
    	TAP_Osd_Delete(menu->rgn);
	    menu->visible = 0;
    }
}

//----------------------------------------------------------------------------
typedef struct  {
    dword revision;
    dword addr;
    int   length;
    byte  from[0x100];
    byte  to[0x100];
}type_patch;

int exTAP_ApplyPatch(type_patch *patch);
int exTAP_RemovePatch(type_patch *patch);


int exTAP_ApplyPatch(type_patch *patch)
{
    int i;

    //check firmware version
    if (patch->revision != _appl_version) {
            TAP_Print("\nonly supports firmware versions 5.12.04,5.12.05,5.12.09");
            TAP_Print("\n");
            return -1;
    }

    //check if firmware is already patched
    for (i=0; i<patch->length; i++) {
        if (*(byte *)(patch->addr+i) == patch->to[i]) {
            if (i == (patch->length-1)) {
                TAP_Print("\nfirmware is already patched - 0x%08x",patch->addr);
                return -1;
            }
        }
        else {
            break;
        }
    }

    //check if original firmware matches expected data
    for (i=0; i<patch->length; i++) {
        if (*(byte *)(patch->addr+i) != patch->from[i]) {
            TAP_Print("\noriginal firmware does not match expected - 0x%08x",patch->addr+i);
            return -2;
        }
    }

    //apply patch
    memcpy((byte *)patch->addr, patch->to, patch->length);
    return 0;
}

int exTAP_RemovePatch(type_patch *patch)
{
    int i;

    //check firmware version
    if (patch->revision != _appl_version) {
            TAP_Print("\nonly supports firmware versions 5.12.04,5.12.05,5.12.09");
            TAP_Print("\n");
            return -1;
    }

    //check if firmware is already patched
    for (i=0; i<patch->length; i++) {
        if (*(byte *)(patch->addr+i) == patch->from[i]) {
            if (i == (patch->length-1)) {
                TAP_Print("\nfirmware is not already patched - 0x%08x",patch->addr);
                return -1;
            }
        }
        else {
            break;
        }
    }

    //check if patched data matches original firmware
    for (i=0; i<patch->length; i++) {
        if (*(byte *)(patch->addr+i) != patch->to[i]) {
            TAP_Print("\npatched data does not match original firmware - 0x%08x",patch->addr+i);
            return -2;
        }
    }

    //apply patch
    memcpy((byte *)patch->addr, patch->from, patch->length);
    return 0;
}

//----------------------------------------------------------------------------
//patch out dsmcc handler
//8007b6e4 0c 01 ea 95   jal   0x8007aa54               # process_dsmcc
//801aacfc 27 bd ff e0   addiu $sp, -0x20               03 e0 00 08    jr   $ra
//801aad00 af b6 00 18   sw    $s6, 0x18($sp)           00 00 00 00    nop

type_patch mheg_off_patch[] = {
    {//12.09
        0x1209,                                             //firmware revision
        0x801aacfc,                                         //patch address
        8,                                                  //length
        {0x27, 0xbd, 0xff, 0xe0, 0xaf, 0xb6, 0x00, 0x18},   //original data
        {0x03, 0xe0, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00},   //new data
    },
    {//12.04
        0x1204,
        0x801ab874,
        8,
        {0x27, 0xbd, 0xff, 0xc8, 0xaf, 0xb3, 0x00, 0x24},
        {0x03, 0xe0, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00},
    },
    {//12.05
        0x1205,
        0x801ab9f8,
        8,
        {0x27, 0xbd, 0xff, 0xc8, 0xaf, 0xb3, 0x00, 0x24},
        {0x03, 0xe0, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00},
    },
};

int patch_dsmcc()
{
    int ret;

    TAP_Print("\nApplying MHEG disable patch");
    ret = exTAP_ApplyPatch(&mheg_off_patch[firmware_id]);
    return ret;
}

int unpatch_dsmcc()
{
    int ret;

    TAP_Print("\nRemoving MHEG disable patch");
    ret = exTAP_RemovePatch(&mheg_off_patch[firmware_id]);
    return ret;
}

//----------------------------------------------------------------------------
